# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import re

from .constants import PYTHON_THREADPOOL_THREAD_COUNT, PYTHON_THREADPOOL_THREAD_COUNT_MIN


class InvalidFileNameError(Exception):

    def __init__(self, file_name: str) -> None:
        super().__init__(
            'Invalid file name: %s', file_name)


def validate_script_file_name(file_name: str):
    # First character can be a letter, number, or underscore
    # Following characters can be a letter, number, underscore, hyphen, or dash
    # Ending must be .py
    pattern = re.compile(r'^[a-zA-Z0-9_][a-zA-Z0-9_\-]*\.py$')
    if not pattern.match(file_name):
        raise InvalidFileNameError(file_name)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import json
import logging

from datetime import datetime
from typing import Any, List, Optional

from .nullable_converters import (
    to_nullable_bool,
    to_nullable_double,
    to_nullable_string,
    to_nullable_timestamp,
)

try:
    from http.cookies import SimpleCookie
except ImportError:
    from Cookie import SimpleCookie


class Datum:
    def __init__(self, value, type):
        self.value = value
        self.type = type

    @property
    def python_value(self) -> Any:
        if self.value is None or self.type is None:
            return None
        elif self.type in ('bytes', 'string', 'int', 'double'):
            return self.value
        elif self.type == 'json':
            return json.loads(self.value)
        elif self.type == 'collection_string':
            return [v for v in self.value.string]
        elif self.type == 'collection_bytes':
            return [v for v in self.value.bytes]
        elif self.type == 'collection_double':
            return [v for v in self.value.double]
        elif self.type == 'collection_sint64':
            return [v for v in self.value.sint64]
        else:
            return self.value

    @property
    def python_type(self) -> type:
        return type(self.python_value)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        return self.value == other.value and self.type == other.type

    def __hash__(self):
        return hash((type(self), (self.value, self.type)))

    def __repr__(self):
        val_repr = repr(self.value)
        if len(val_repr) > 10:
            val_repr = val_repr[:10] + '...'
        return '<Datum ' + str(self.type) + val_repr + '>'

    @classmethod
    def from_typed_data(cls, protos):
        try:
            td = protos.TypedData
        except Exception:
            # Todo: better catch for Datum.from_typed_data(http.body)
            #  if the data being sent in is already protos.TypedData
            td = protos
        tt = td.WhichOneof('data')
        if tt == 'http':
            http = td.http
            val = dict(
                method=Datum(http.method, 'string'),
                url=Datum(http.url, 'string'),
                headers={
                    k: Datum(v, 'string') for k, v in http.headers.items()
                },
                body=(
                    Datum.from_typed_data(http.body)
                    or Datum(type='bytes', value=b'')
                ),
                params={
                    k: Datum(v, 'string') for k, v in http.params.items()
                },
                query={
                    k: Datum(v, 'string') for k, v in http.query.items()
                },
            )
        elif tt == 'string':
            val = td.string
        elif tt == 'bytes':
            val = td.bytes
        elif tt == 'json':
            val = td.json
        elif tt == 'collection_bytes':
            val = td.collection_bytes
        elif tt == 'collection_string':
            val = td.collection_string
        elif tt == 'collection_sint64':
            val = td.collection_sint64
        elif tt == 'model_binding_data':
            val = td.model_binding_data
        elif tt is None:
            return None
        else:
            raise NotImplementedError(
                'unsupported TypeData kind: {!r}'.format(tt)
            )

        return cls(val, tt)


def datum_as_proto(datum: Datum, protos):
    if datum.type == 'string':
        return protos.TypedData(string=datum.value)
    elif datum.type == 'bytes':
        return protos.TypedData(bytes=datum.value)
    elif datum.type == 'json':
        return protos.TypedData(json=datum.value)
    elif datum.type == 'http':
        return protos.TypedData(http=protos.RpcHttp(
            status_code=datum.value['status_code'].value,
            headers={
                k: v.value
                for k, v in datum.value['headers'].items()
            },
            cookies=parse_to_rpc_http_cookie_list(datum.value.get('cookies'), protos),
            enable_content_negotiation=False,
            body=datum_as_proto(datum.value['body'], protos),
        ))
    elif datum.type is None:
        return None
    elif datum.type == 'dict':
        # TypedData doesn't support dict, so we return it as json
        return protos.TypedData(json=json.dumps(datum.value))
    elif datum.type == 'list':
        # TypedData doesn't support list, so we return it as json
        return protos.TypedData(json=json.dumps(datum.value))
    elif datum.type == 'int':
        return protos.TypedData(int=datum.value)
    elif datum.type == 'double':
        return protos.TypedData(double=datum.value)
    elif datum.type == 'bool':
        # TypedData doesn't support bool, so we return it as an int
        return protos.TypedData(int=int(datum.value))
    else:
        raise NotImplementedError(
            'unexpected Datum type: {!r}'.format(datum.type)
        )


def parse_to_rpc_http_cookie_list(cookies: Optional[List[SimpleCookie]], protos):
    if cookies is None:
        return cookies

    rpc_http_cookies = []

    for cookie in cookies:
        for name, cookie_entity in cookie.items():
            rpc_http_cookies.append(
                protos.RpcHttpCookie(name=name,
                                     value=cookie_entity.value,
                                     domain=to_nullable_string(
                                         cookie_entity['domain'],
                                         'cookie.domain',
                                         protos),
                                     path=to_nullable_string(
                                         cookie_entity['path'],
                                         'cookie.path',
                                         protos),
                                     expires=to_nullable_timestamp(
                                         parse_cookie_attr_expires(
                                             cookie_entity), 'cookie.expires',
                                         protos),
                                     secure=to_nullable_bool(
                                         bool(cookie_entity['secure']),
                                         'cookie.secure',
                                         protos),
                                     http_only=to_nullable_bool(
                                         bool(cookie_entity['httponly']),
                                         'cookie.httpOnly',
                                         protos),
                                     same_site=parse_cookie_attr_same_site(
                                         cookie_entity, protos),
                                     max_age=to_nullable_double(
                                         cookie_entity['max-age'],
                                         'cookie.maxAge',
                                         protos)))

    return rpc_http_cookies


def parse_cookie_attr_expires(cookie_entity):
    expires = cookie_entity['expires']

    if expires is not None and len(expires) != 0:
        try:
            return datetime.strptime(expires, "%a, %d %b %Y %H:%M:%S GMT")
        except ValueError:
            logging.error(
                "Can not parse value %s of expires in the cookie "
                "due to invalid format.", expires)
            raise
        except OverflowError:
            logging.error(
                "Can not parse value %s of expires in the cookie "
                "because the parsed date exceeds the largest valid C "
                "integer on your system.", expires)
            raise

    return None


def parse_cookie_attr_same_site(cookie_entity, protos):
    same_site = getattr(protos.RpcHttpCookie.SameSite, "None")
    try:
        raw_same_site_str = cookie_entity['samesite'].lower()

        if raw_same_site_str == 'lax':
            same_site = protos.RpcHttpCookie.SameSite.Lax
        elif raw_same_site_str == 'strict':
            same_site = protos.RpcHttpCookie.SameSite.Strict
        elif raw_same_site_str == 'none':
            same_site = protos.RpcHttpCookie.SameSite.ExplicitNone
    except Exception:
        return same_site

    return same_site



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
# TODO: organize this better

import sys

TRUE = "true"
TRACEPARENT = "traceparent"
TRACESTATE = "tracestate"

# Capabilities
RAW_HTTP_BODY_BYTES = "RawHttpBodyBytes"
TYPED_DATA_COLLECTION = "TypedDataCollection"
RPC_HTTP_BODY_ONLY = "RpcHttpBodyOnly"
RPC_HTTP_TRIGGER_METADATA_REMOVED = "RpcHttpTriggerMetadataRemoved"
WORKER_STATUS = "WorkerStatus"
SHARED_MEMORY_DATA_TRANSFER = "SharedMemoryDataTransfer"
FUNCTION_DATA_CACHE = "FunctionDataCache"
HTTP_URI = "HttpUri"
REQUIRES_ROUTE_PARAMETERS = "RequiresRouteParameters"
# When this capability is enabled, logs are not piped back to the
# host from the worker. Logs will directly go to where the user has
# configured them to go. This is to ensure that the logs are not
# duplicated.
WORKER_OPEN_TELEMETRY_ENABLED = "WorkerOpenTelemetryEnabled"

# Platform Environment Variables
AZURE_WEBJOBS_SCRIPT_ROOT = "AzureWebJobsScriptRoot"
CONTAINER_NAME = "CONTAINER_NAME"

# Python Specific Feature Flags and App Settings
PYTHON_THREADPOOL_THREAD_COUNT = "PYTHON_THREADPOOL_THREAD_COUNT"
PYTHON_ENABLE_DEBUG_LOGGING = "PYTHON_ENABLE_DEBUG_LOGGING"
# Appsetting to turn on OpenTelemetry support/features
# Includes turning on Azure monitor distro to send telemetry to AppInsights
PYTHON_ENABLE_OPENTELEMETRY = "PYTHON_ENABLE_OPENTELEMETRY"
# Appsetting to specify root logger name of logger to collect telemetry for
# Used by Azure monitor distro
PYTHON_AZURE_MONITOR_LOGGER_NAME = "PYTHON_AZURE_MONITOR_LOGGER_NAME"

# Setting Defaults
PYTHON_THREADPOOL_THREAD_COUNT_DEFAULT = 1
PYTHON_THREADPOOL_THREAD_COUNT_MIN = 1
PYTHON_THREADPOOL_THREAD_COUNT_MAX = sys.maxsize
PYTHON_THREADPOOL_THREAD_COUNT_MAX_37 = 32
PYTHON_ENABLE_OPENTELEMETRY_DEFAULT = False
PYTHON_AZURE_MONITOR_LOGGER_NAME_DEFAULT = ""

# External Site URLs
MODULE_NOT_FOUND_TS_URL = "https://aka.ms/functions-modulenotfound"

PYTHON_LANGUAGE_RUNTIME = "python"

# Paths
CUSTOMER_PACKAGES_PATH = "/home/site/wwwroot/.python_packages/lib/site" \
                         "-packages"

METADATA_PROPERTIES_WORKER_INDEXED = "worker_indexed"

# Header names
X_MS_INVOCATION_ID = "x-ms-invocation-id"

# Trigger Names
HTTP_TRIGGER = "httpTrigger"

# Output Names
HTTP = "http"

# Appsetting to specify AppInsights connection string
APPLICATIONINSIGHTS_CONNECTION_STRING = "APPLICATIONINSIGHTS_CONNECTION_STRING"



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Dict


class TraceContext:
    """Check https://www.w3.org/TR/trace-context/ for more information"""

    def __init__(self, trace_parent: str,
                 trace_state: str, attributes: Dict[str, str]) -> None:
        self.__trace_parent = trace_parent
        self.__trace_state = trace_state
        self.__attributes = attributes

    @property
    def Tracestate(self) -> str:
        """Get trace state from trace-context (deprecated)."""
        return self.__trace_state

    @property
    def Traceparent(self) -> str:
        """Get trace parent from trace-context (deprecated)."""
        return self.__trace_parent

    @property
    def Attributes(self) -> Dict[str, str]:
        """Get trace-context attributes (deprecated)."""
        return self.__attributes

    @property
    def trace_state(self) -> str:
        """Get trace state from trace-context"""
        return self.__trace_state

    @property
    def trace_parent(self) -> str:
        """Get trace parent from trace-context"""
        return self.__trace_parent

    @property
    def attributes(self) -> Dict[str, str]:
        """Get trace-context attributes"""
        return self.__attributes



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import os
import platform
import sys
from .constants import PYTHON_LANGUAGE_RUNTIME
from ..logging import logger
from ..version import VERSION


def change_cwd(new_cwd: str):
    if os.path.exists(new_cwd):
        os.chdir(new_cwd)
        logger.info('Changing current working directory to %s', new_cwd)
    else:
        logger.warning('Directory %s is not found when reloading', new_cwd)


def get_worker_metadata(protos):
    return protos.WorkerMetadata(
        runtime_name=PYTHON_LANGUAGE_RUNTIME,
        runtime_version=str(sys.version_info.major) + "." + str(sys.version_info.minor),
        worker_version=VERSION,
        worker_bitness=platform.machine(),
        custom_properties={})



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import threading

from typing import Type

from .retrycontext import RetryContext
from .tracecontext import TraceContext


class Context:
    def __init__(self,
                 func_name: str,
                 func_dir: str,
                 invocation_id: str,
                 thread_local_storage: Type[threading.local],
                 trace_context: TraceContext,
                 retry_context: RetryContext) -> None:
        self.__func_name = func_name
        self.__func_dir = func_dir
        self.__invocation_id = invocation_id
        self.__thread_local_storage = thread_local_storage
        self.__trace_context = trace_context
        self.__retry_context = retry_context

    @property
    def invocation_id(self) -> str:
        return self.__invocation_id

    @property
    def thread_local_storage(self) -> Type[threading.local]:
        return self.__thread_local_storage

    @property
    def function_name(self) -> str:
        return self.__func_name

    @property
    def function_directory(self) -> str:
        return self.__func_dir

    @property
    def trace_context(self) -> TraceContext:
        return self.__trace_context

    @property
    def retry_context(self) -> RetryContext:
        return self.__retry_context


def get_context(invoc_request, name: str,
                directory: str) -> Context:
    """ For more information refer:
    https://aka.ms/azfunc-invocation-context
    """
    trace_context = TraceContext(
        invoc_request.trace_context.trace_parent,
        invoc_request.trace_context.trace_state,
        invoc_request.trace_context.attributes)

    retry_context = RetryContext(
        invoc_request.retry_context.retry_count,
        invoc_request.retry_context.max_retry_count,
        invoc_request.retry_context.exception)

    return Context(
        name, directory, invoc_request.invocation_id,
        threading.local(), trace_context, retry_context)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import logging
import os
import sys

from typing import List, Optional

from .functions import FunctionInfo, Registry
from .loader import load_function, install
from .logging import logger
from .otel import otel_manager, initialize_azure_monitor, configure_opentelemetry

from .bindings.context import get_context
from .bindings.meta import load_binding_registry, is_trigger_binding, from_incoming_proto, to_outgoing_param_binding, to_outgoing_proto
from .bindings.out import Out
from .utils.constants import (FUNCTION_DATA_CACHE,
                              RAW_HTTP_BODY_BYTES,
                              TYPED_DATA_COLLECTION,
                              RPC_HTTP_BODY_ONLY,
                              WORKER_STATUS,
                              RPC_HTTP_TRIGGER_METADATA_REMOVED,
                              SHARED_MEMORY_DATA_TRANSFER,
                              TRUE,
                              PYTHON_ENABLE_OPENTELEMETRY,
                              PYTHON_ENABLE_OPENTELEMETRY_DEFAULT,
                              WORKER_OPEN_TELEMETRY_ENABLED,
                              PYTHON_ENABLE_DEBUG_LOGGING)
from .utils.current import get_current_loop, execute_async, run_sync_func
from .utils.env_state import get_app_setting, is_envvar_true
from .utils.helpers import change_cwd, get_worker_metadata
from .utils.tracing import serialize_exception

_functions = Registry()
_function_data_cache_enabled: bool = False
_host: str = None
protos = None


# Protos will be the retry / binding / metadata protos object that we populate and return
async def worker_init_request(request):
    logger.info("V1 Library Worker: received worker_init_request")
    global _host, protos, _function_data_cache_enabled
    init_request = request.request.worker_init_request
    host_capabilities = init_request.capabilities
    _host = request.properties.get("host")
    protos = request.properties.get("protos")
    if FUNCTION_DATA_CACHE in host_capabilities:
        val = host_capabilities[FUNCTION_DATA_CACHE]
        _function_data_cache_enabled = val == TRUE

    capabilities = {
        RAW_HTTP_BODY_BYTES: TRUE,
        TYPED_DATA_COLLECTION: TRUE,
        RPC_HTTP_BODY_ONLY: TRUE,
        WORKER_STATUS: TRUE,
        RPC_HTTP_TRIGGER_METADATA_REMOVED: TRUE,
        SHARED_MEMORY_DATA_TRANSFER: TRUE,
    }
    if get_app_setting(setting=PYTHON_ENABLE_OPENTELEMETRY,
                       default_value=PYTHON_ENABLE_OPENTELEMETRY_DEFAULT):
        initialize_azure_monitor()

        if otel_manager.get_azure_monitor_available():
            capabilities[WORKER_OPEN_TELEMETRY_ENABLED] = TRUE


    # loading bindings registry and saving results to a static
    # dictionary which will be later used in the invocation request
    load_binding_registry()
    install()

    # No indexing in init for V1 functions

    return protos.WorkerInitResponse(
        capabilities=capabilities,
        worker_metadata=get_worker_metadata(protos),
        result=protos.StatusResult(status=protos.StatusResult.Success)
    )


# worker_status_request can be done in the proxy worker

async def functions_metadata_request(request):
    logger.info("V1 Library Worker: received worker_metadata_request")
    global protos
    # Setting result as None here. If this is a V1 function, nothing to return
    return protos.FunctionMetadataResponse(
                    use_default_metadata_indexing=True,
                    function_metadata_results=None,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Success))


async def function_load_request(request):
    logger.info("V1 Library Worker: received worker_load_request")
    global protos
    func_request = request.request.function_load_request
    function_id = func_request.function_id
    function_metadata = func_request.metadata
    function_name = function_metadata.name
    function_app_directory = function_metadata.directory

    try:
        if not _functions.get_function(function_id):
            func = load_function(
                function_name,
                function_app_directory,
                func_request.metadata.script_file,
                func_request.metadata.entry_point)

            _functions.add_function(
                function_id, func, func_request.metadata, protos)

        return protos.FunctionLoadResponse(
                    function_id=function_id,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Success))

    except Exception as ex:
        return protos.FunctionLoadResponse(
                    function_id=function_id,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Failure,
                        exception=serialize_exception(ex, protos)))


async def invocation_request(request):
    logger.info("Library Worker: received worker_invocation_request")
    global protos
    invoc_request = request.request.invocation_request
    invocation_id = invoc_request.invocation_id
    function_id = invoc_request.function_id
    threadpool = request.properties.get("threadpool")

    try:
        fi: FunctionInfo = _functions.get_function(
            function_id)
        assert fi is not None

        args = {}

        for pb in invoc_request.input_data:
            pb_type_info = fi.input_types[pb.name]
            if is_trigger_binding(pb_type_info.binding_name):
                trigger_metadata = invoc_request.trigger_metadata
            else:
                trigger_metadata = None

            args[pb.name] = from_incoming_proto(
                pb_type_info.binding_name,
                pb,
                trigger_metadata=trigger_metadata)

        fi_context = get_context(invoc_request, fi.name,
                                 fi.directory)

        # Use local thread storage to store the invocation ID
        # for a customer's threads
        fi_context.thread_local_storage.invocation_id = invocation_id
        if fi.requires_context:
            args['context'] = fi_context

        if fi.output_types:
            for name in fi.output_types:
                args[name] = Out()

        if fi.is_async:
            if otel_manager.get_azure_monitor_available():
                configure_opentelemetry(fi_context)

            call_result = await execute_async(fi.func, args)  # Not supporting Extensions
        else:
            _loop = get_current_loop()
            call_result = await _loop.run_in_executor(
                threadpool,
                run_sync_func,
                invocation_id, fi_context, fi.func, args)

        if call_result is not None and not fi.has_return:
            raise RuntimeError(
                'function %s without a $return binding'
                'returned a non-None value', repr(fi.name))

        output_data = []
        if fi.output_types:
            for out_name, out_type_info in fi.output_types.items():
                val = args[out_name].get()
                if val is None:
                    # TODO: is the "Out" parameter optional?
                    # Can "None" be marshaled into protos.TypedData?
                    continue

                param_binding = to_outgoing_param_binding(
                    out_type_info.binding_name, val,
                    pytype=out_type_info.pytype,
                    out_name=out_name,
                    protos=protos)
                output_data.append(param_binding)

        return_value = None
        if fi.return_type is not None:
            return_value = to_outgoing_proto(
                fi.return_type.binding_name,
                call_result,
                pytype=fi.return_type.pytype,
                protos=protos
            )

        # Actively flush customer print() function to console
        sys.stdout.flush()

        return protos.InvocationResponse(
                    invocation_id=invocation_id,
                    return_value=return_value,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Success),
                    output_data=output_data)

    except Exception as ex:
        return protos.InvocationResponse(
                    invocation_id=invocation_id,
                    result=protos.StatusResult(
                        status=protos.StatusResult.Failure,
                        exception=serialize_exception(ex, protos)))


async def function_environment_reload_request(request):
    """Only runs on Linux Consumption placeholder specialization.
    This is called only when placeholder mode is true. On worker restarts
    worker init request will be called directly.
    """
    logger.info("V1 Library Worker: received worker_env_reload_request")
    try:
        global protos

        func_env_reload_request = \
            request.request.function_environment_reload_request
        protos = request.properties.get("protos")

        # Append function project root to module finding sys.path
        if func_env_reload_request.function_app_directory:
            sys.path.append(func_env_reload_request.function_app_directory)

        # Clear sys.path import cache, reload all module from new sys.path
        sys.path_importer_cache.clear()

        # Reload environment variables
        os.environ.clear()
        env_vars = func_env_reload_request.environment_variables
        for var in env_vars:
            os.environ[var] = env_vars[var]

        if is_envvar_true(PYTHON_ENABLE_DEBUG_LOGGING):
            root_logger = logging.getLogger("azure.functions")
            root_logger.setLevel(logging.DEBUG)

        # calling load_binding_registry again since the
        # reload_customer_libraries call clears the registry
        load_binding_registry()
        install()

        capabilities = {}
        if get_app_setting(
                setting=PYTHON_ENABLE_OPENTELEMETRY,
                default_value=PYTHON_ENABLE_OPENTELEMETRY_DEFAULT):
            initialize_azure_monitor()

            if otel_manager.get_azure_monitor_available():
                capabilities[WORKER_OPEN_TELEMETRY_ENABLED] = (
                    TRUE)



        # Change function app directory
        if getattr(func_env_reload_request,
                   'function_app_directory', None):
            change_cwd(
                func_env_reload_request.function_app_directory)

        return protos.FunctionEnvironmentReloadResponse(
                capabilities=capabilities,
                worker_metadata=get_worker_metadata(protos),
                result=protos.StatusResult(
                    status=protos.StatusResult.Success))

    except Exception as ex:
        return protos.FunctionEnvironmentReloadResponse(
                result=protos.StatusResult(
                    status=protos.StatusResult.Failure,
                    exception=serialize_exception(ex, protos)))



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import importlib
import importlib.machinery
import os
import os.path
import pathlib
import sys
from os import PathLike, fspath
from typing import Optional

from .utils.constants import (
    CUSTOMER_PACKAGES_PATH,
    MODULE_NOT_FOUND_TS_URL,
)
from .utils.wrappers import attach_message_to_exception

_AZURE_NAMESPACE = '__app__'
_DEFAULT_SCRIPT_FILENAME = '__init__.py'
_DEFAULT_ENTRY_POINT = 'main'
_submodule_dirs = []


def register_function_dir(path: PathLike) -> None:
    try:
        _submodule_dirs.append(fspath(path))
    except TypeError as e:
        raise RuntimeError('Path (%s) is incompatible with fspath. '
                           'It is of type %s.', path, type(path), e)


def install() -> None:
    if _AZURE_NAMESPACE not in sys.modules:
        # Create and register the __app__ namespace package.
        ns_spec = importlib.machinery.ModuleSpec(_AZURE_NAMESPACE, None)
        ns_spec.submodule_search_locations = _submodule_dirs
        ns_pkg = importlib.util.module_from_spec(ns_spec)
        sys.modules[_AZURE_NAMESPACE] = ns_pkg


@attach_message_to_exception(
    expt_type=ImportError,
    message="Cannot find module. Please check the requirements.txt file for the missing module. For more info, please refer the troubleshooting guide: " + MODULE_NOT_FOUND_TS_URL +
            ". Current sys.path: " + " ".join(sys.path),
    debug_logs="Error in index_function_app. Sys Path:" + " ".join(sys.path) +
               ", python-packages Path exists: " + str(os.path.exists(CUSTOMER_PACKAGES_PATH)))
def load_function(name: str, directory: str, script_file: str,
                  entry_point: Optional[str]):
    dir_path = pathlib.Path(directory)
    script_path = pathlib.Path(script_file) if script_file else pathlib.Path(
        _DEFAULT_SCRIPT_FILENAME)
    if not entry_point:
        entry_point = _DEFAULT_ENTRY_POINT

    register_function_dir(dir_path.parent)

    try:
        rel_script_path = script_path.relative_to(dir_path.parent)
    except ValueError:
        raise RuntimeError(
            'script path %s is not relative to the specified '
            'directory %s', script_file, directory
        )

    last_part = rel_script_path.parts[-1]
    modname, ext = os.path.splitext(last_part)
    if ext != '.py':
        raise RuntimeError(
            'cannot load function %s: '
            'invalid Python filename %s', name, script_file)

    modname_parts = [_AZURE_NAMESPACE]
    modname_parts.extend(rel_script_path.parts[:-1])

    # If the __init__.py contains the code, we should avoid double loading.
    if modname.lower() != '__init__':
        modname_parts.append(modname)

    fullmodname = '.'.join(modname_parts)

    mod = importlib.import_module(fullmodname)

    func = getattr(mod, entry_point, None)
    if func is None or not callable(func):
        raise RuntimeError(
            'cannot load function %s: function %s() is not '
            'present in %s', name, entry_point, rel_script_path)

    return func






# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from typing import Any, Callable

from ..logging import logger
from .env_state import is_envvar_false, is_envvar_true
from .tracing import extend_exception_message


def enable_feature_by(flag: str,
                      default: Any = None,
                      flag_default: bool = False) -> Callable:
    def decorate(func):
        def call(*args, **kwargs):
            if is_envvar_true(flag):
                return func(*args, **kwargs)
            if flag_default and not is_envvar_false(flag):
                return func(*args, **kwargs)
            return default
        return call
    return decorate


def disable_feature_by(flag: str,
                       default: Any = None,
                       flag_default: bool = False) -> Callable:
    def decorate(func):
        def call(*args, **kwargs):
            if is_envvar_true(flag):
                return default
            if flag_default and not is_envvar_false(flag):
                return default
            return func(*args, **kwargs)
        return call
    return decorate


def attach_message_to_exception(expt_type: Exception, message: str,
                                debug_logs=None) -> Callable:
    def decorate(func):
        def call(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except expt_type as e:
                if debug_logs is not None:
                    logger.error(debug_logs)
                raise extend_exception_message(e, message)
        return call
    return decorate



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

class Out:

    def __init__(self) -> None:
        self.__value = None

    def set(self, val):
        self.__value = val

    def get(self) -> str:
        return self.__value



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
"""Unittest helpers.

All functions in this file should be considered private APIs,
and can be changed without a notice.
"""

import asyncio
import functools
import inspect
import unittest


class AsyncTestCaseMeta(type(unittest.TestCase)):
    def __new__(mcls, name, bases, ns):
        for attrname, attr in ns.items():
            if (attrname.startswith('test_')
                    and inspect.iscoroutinefunction(attr)):
                ns[attrname] = mcls._sync_wrap(attr)

        return super().__new__(mcls, name, bases, ns)

    @staticmethod
    def _sync_wrap(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return asyncio.run(func(*args, **kwargs))

        return wrapper


class AsyncTestCase(unittest.TestCase, metaclass=AsyncTestCaseMeta):
    pass



from .FunctionRpc_pb2_grpc import (  # NoQA
    FunctionRpcStub,
    FunctionRpcServicer,
    add_FunctionRpcServicer_to_server)

from .FunctionRpc_pb2 import (  # NoQA
    StreamingMessage,
    StartStream,
    WorkerInitRequest,
    WorkerInitResponse,
    RpcFunctionMetadata,
    FunctionLoadRequest,
    FunctionLoadResponse,
    FunctionEnvironmentReloadRequest,
    FunctionEnvironmentReloadResponse,
    InvocationRequest,
    InvocationResponse,
    WorkerHeartbeat,
    WorkerStatusRequest,
    WorkerStatusResponse,
    BindingInfo,
    StatusResult,
    RpcException,
    ParameterBinding,
    TypedData,
    RpcHttp,
    RpcHttpCookie,
    RpcLog,
    RpcSharedMemory,
    RpcDataType,
    CloseSharedMemoryResourcesRequest,
    CloseSharedMemoryResourcesResponse,
    FunctionsMetadataRequest,
    FunctionMetadataResponse,
    WorkerMetadata,
    RpcRetryOptions)

from .shared.NullableTypes_pb2 import (
    NullableString,
    NullableBool,
    NullableDouble,
    NullableTimestamp
)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import unittest

from typing import Any

from tests.utils import testutils
import tests.protos as protos

from azure_functions_worker_v1.handle_event import (worker_init_request,
                                                 functions_metadata_request,
                                                 function_load_request,
                                                 function_environment_reload_request)


class WorkerRequest:
    def __init__(self, name: str, request: Any, properties: dict):
        self.name = name
        self.request = request
        self.properties = properties


class InnerRequest:
    def __init__(self, name: Any):
        self.worker_init_request = name
        self.function_load_request = name
        self.function_environment_reload_request = name


class InnerInnerRequest:
    def __init__(self, name: Any):
        self.capabilities = name
        self.function_app_directory = "C:\\Users\\victoriahall\\Documents\\repos\\azure-functions-python-worker-313\\tests\\unittests\\default_template"
        self.function_id = 1
        self.metadata = Metadata(name="default_template")
        self.environment_variables = {}


class Metadata:
    def __init__(self, name: Any):
        self.name = name
        self.directory = 'C:\\Users\\victoriahall\\Documents\\repos\\basic-http-v1\\HttpTrigger1'
        self.script_file = 'C:\\Users\\victoriahall\\Documents\\repos\\basic-http-v1\\HttpTrigger1\\__init__.py'
        self.entry_point = ""
        self.bindings = {'req': Binding("httpTrigger"), '$return': Binding("http", 1)}
        self.environment_variables = {}


class Binding:
    def __init__(self, type, direction=""):
        self.type = type
        self.direction = direction


class TestObjects(unittest.TestCase):
    def test_stringify_enum(self):
        pass

    def test_status(self):
        pass

    def test_worker_response(self):
        pass


class TestHandleEvent(testutils.AsyncTestCase):
    async def test_worker_init_request(self):
        worker_request = WorkerRequest(name='worker_init_request',
                                       request=InnerRequest(InnerInnerRequest('hello')),
                                       properties={'host': '123',
                                                   'protos': protos})
        result = await worker_init_request(worker_request)
        self.assertEqual(result.capabilities,
                         {'WorkerStatus': 'true', 'RpcHttpBodyOnly': 'true', 'SharedMemoryDataTransfer': 'true',
                          'RpcHttpTriggerMetadataRemoved': 'true', 'RawHttpBodyBytes': 'true',
                          'TypedDataCollection': 'true'})
        self.assertEqual(result.worker_metadata.runtime_name, "python")
        self.assertIsNotNone(result.worker_metadata.runtime_version)
        self.assertIsNotNone(result.worker_metadata.worker_version)
        self.assertIsNotNone(result.worker_metadata.worker_bitness)
        self.assertEqual(result.result.status, 1)

    def test_worker_init_request_with_streaming(self):
        pass

    def test_worker_init_request_with_exception(self):
        pass

    async def test_functions_metadata_request(self):
        worker_request2 = WorkerRequest(name='worker_init_request',
                                       request=InnerRequest(InnerInnerRequest('hello')),
                                       properties={'host': '123',
                                                   'protos': protos})
        result2 = await worker_init_request(worker_request2)
        worker_request = WorkerRequest(name='functions_metadata_request',
                                       request=InnerRequest(InnerInnerRequest('hello')),
                                       properties={})
        result = await functions_metadata_request(worker_request)
        self.assertEqual(result.use_default_metadata_indexing, True)
        self.assertEqual(result.function_metadata_results, [])
        self.assertEqual(result.result.status, 1)

    def test_functions_metadata_request_with_exception(self):
        pass

    async def test_functions_load_request(self):
        worker_request = WorkerRequest(name='functions_load_request',
                                       request=InnerRequest(InnerInnerRequest('hello')),
                                       properties={'protos': protos})
        init_result = await worker_init_request(worker_request)
        result = await function_load_request(worker_request)
        self.assertEqual(result.name, "functions_load_request")

    def test_invocation_request_sync(self):
        pass

    def test_invocation_request_async(self):
        pass

    def test_invocation_request_with_exception(self):
        pass

    async def test_function_environment_reload_request(self):
        worker_request = WorkerRequest(name='function_environment_reload_request',
                                       request=InnerRequest(InnerInnerRequest('hello')),
                                       properties={'host': '123',
                                                   'protos': protos})
        result = await function_environment_reload_request(worker_request)
        self.assertEqual(result.name, "function_environment_reload_request")
        self.assertEqual(result.status, Status.SUCCESS)
        self.assertEqual(result.result.get('capabilities'), {})
        self.assertEqual(result.exception, None)

    def test_function_environment_reload_request_with_streaming(self):
        pass

    def test_function_environment_reload_request_with_exception(self):
        pass

    def test_load_function_metadata(self):
        pass

    def test_index_functions(self):
        pass



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import asyncio
import functools

from typing import Any

from ..otel import otel_manager, configure_opentelemetry


def get_current_loop():
    return asyncio.events.get_event_loop()


async def execute_async(function, args) -> Any:
    return await function(**args)

def execute_sync(function, args) -> Any:
    return function(**args)


def run_sync_func(invocation_id, context, func, params):
    # This helper exists because we need to access the current
    # invocation_id from ThreadPoolExecutor's threads.
    context.thread_local_storage.invocation_id = invocation_id
    try:
        if otel_manager.get_azure_monitor_available():
            configure_opentelemetry(context)
        result = functools.partial(execute_sync, func)
        return result(params)
    finally:
        context.thread_local_storage.invocation_id = None




# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import os
import sys
from typing import Callable, Optional


def is_true_like(setting: str) -> bool:
    if setting is None:
        return False

    return setting.lower().strip() in {'1', 'true', 't', 'yes', 'y'}


def is_false_like(setting: str) -> bool:
    if setting is None:
        return False

    return setting.lower().strip() in {'0', 'false', 'f', 'no', 'n'}


def is_envvar_true(env_key: str) -> bool:
    if os.getenv(env_key) is None:
        return False

    return is_true_like(os.environ[env_key])


def is_envvar_false(env_key: str) -> bool:
    if os.getenv(env_key) is None:
        return False

    return is_false_like(os.environ[env_key])


def get_app_setting(
    setting: str,
    default_value: Optional[str] = None,
    validator: Optional[Callable[[str], bool]] = None
) -> Optional[str]:
    """Returns the application setting from environment variable.

    Parameters
    ----------
    setting: str
        The name of the application setting (e.g. FUNCTIONS_RUNTIME_VERSION)

    default_value: Optional[str]
        The expected return value when the application setting is not found,
        or the app setting does not pass the validator.

    validator: Optional[Callable[[str], bool]]
        A function accepts the app setting value and should return True when
        the app setting value is acceptable.

    Returns
    -------
    Optional[str]
        A string value that is set in the application setting
    """
    app_setting_value = os.getenv(setting)

    # If an app setting is not configured, we return the default value
    if app_setting_value is None:
        return default_value

    # If there's no validator, we should return the app setting value directly
    if validator is None:
        return app_setting_value

    # If the app setting is set with a validator,
    # On True, should return the app setting value
    # On False, should return the default value
    if validator(app_setting_value):
        return app_setting_value
    return default_value



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import typing
from typing import Any, Optional

from .datumdef import Datum


class GenericBinding:

    @classmethod
    def has_trigger_support(cls) -> bool:
        return False

    @classmethod
    def check_input_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, (str, bytes))

    @classmethod
    def check_output_type_annotation(cls, pytype: type) -> bool:
        return issubclass(pytype, (str, bytes, bytearray))

    @classmethod
    def encode(cls, obj: Any, *,
               expected_type: Optional[type]) -> Datum:
        if isinstance(obj, str):
            return Datum(type='string', value=obj)

        elif isinstance(obj, (bytes, bytearray)):
            return Datum(type='bytes', value=bytes(obj))
        elif obj is None:
            return Datum(type=None, value=obj)
        elif isinstance(obj, dict):
            return Datum(type='dict', value=obj)
        elif isinstance(obj, list):
            return Datum(type='list', value=obj)
        elif isinstance(obj, int):
            return Datum(type='int', value=obj)
        elif isinstance(obj, float):
            return Datum(type='double', value=obj)
        elif isinstance(obj, bool):
            return Datum(type='bool', value=obj)
        else:
            raise NotImplementedError

    @classmethod
    def decode(cls, data: Datum, *, trigger_metadata) -> typing.Any:
        # Enabling support for Dapr bindings
        # https://github.com/Azure/azure-functions-python-worker/issues/1316
        if data is None:
            return None
        data_type = data.type

        if data_type == 'string':
            result = data.value
        elif data_type == 'bytes':
            result = data.value
        elif data_type == 'json':
            result = data.value
        elif data_type is None:
            result = None
        else:
            raise ValueError(
                'unexpected type of data received for the "generic" binding ', repr(data_type)
            )

        return result

    @classmethod
    def has_implicit_output(cls, bind_name: Optional[str]) -> bool:
        if bind_name == 'durableClient':
            return False
        return True



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import traceback

from traceback import StackSummary, extract_tb
from typing import List


def extend_exception_message(exc: Exception, msg: str) -> Exception:
    # Reconstruct exception message
    # From: ImportModule: no module name
    #   To: ImportModule: no module name. msg
    old_tb = exc.__traceback__
    old_msg = getattr(exc, 'msg', None) or str(exc) or ''
    new_msg = (old_msg.rstrip('.') + '. ' + msg).rstrip()
    new_excpt = type(exc)(new_msg).with_traceback(old_tb)
    return new_excpt


def marshall_exception_trace(exc: Exception) -> str:
    stack_summary: StackSummary = extract_tb(exc.__traceback__)
    if isinstance(exc, ModuleNotFoundError):
        stack_summary = _marshall_module_not_found_error(stack_summary)
    return ''.join(stack_summary.format())


def _marshall_module_not_found_error(tbss: StackSummary) -> StackSummary:
    tbss = _remove_frame_from_stack(tbss, '<frozen importlib._bootstrap>')
    tbss = _remove_frame_from_stack(
        tbss, '<frozen importlib._bootstrap_external>')
    return tbss


def _remove_frame_from_stack(tbss: StackSummary,
                             framename: str) -> StackSummary:
    filtered_stack_list: List[traceback.FrameSummary] = \
        list(filter(lambda frame: getattr(frame,
                                          'filename') != framename, tbss))
    filtered_stack: StackSummary = StackSummary.from_list(filtered_stack_list)
    return filtered_stack


def serialize_exception(exc: Exception, protos):
    try:
        message = str(type(exc).__name__) + ": " + str(exc)
    except Exception:
        message = ('Unhandled exception in function. '
                   'Could not serialize original exception message.')
    try:
        stack_trace = marshall_exception_trace(exc)
    except Exception:
        stack_trace = ''
    return protos.RpcException(message=message, stack_trace=stack_trace)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from datetime import datetime
from typing import Optional, Union


def to_nullable_string(nullable: Optional[str], property_name: str, protos):
    """Converts string input to an 'NullableString' to be sent through the
    RPC layer. Input that is not a string but is also not null or undefined
    logs a function app level warning.

    :param nullable Input to be converted to an NullableString if it is a
    valid string
    :param property_name The name of the property that the caller will
    assign the output to. Used for debugging.
    :param: protos The protos object used for returning the appropriate value
    """
    if isinstance(nullable, str):
        return protos.NullableString(value=nullable)

    if nullable is not None:
        raise TypeError(
            "A 'str' type was expected instead of a '%s' "
            "type. Cannot parse value %s of '%s'.", type(nullable), nullable, property_name)

    return None


def to_nullable_bool(nullable: Optional[bool], property_name: str, protos):
    """Converts boolean input to an 'NullableBool' to be sent through the
    RPC layer. Input that is not a boolean but is also not null or undefined
    logs a function app level warning.

    :param nullable Input to be converted to an NullableBool if it is a
    valid boolean
    :param property_name The name of the property that the caller will
    assign the output to. Used for debugging.
    :param protos The protos object used for returning the appropriate value
    """
    if isinstance(nullable, bool):
        return protos.NullableBool(value=nullable)

    if nullable is not None:
        raise TypeError(
            "A 'bool' type was expected instead of a '%s' "
            "type. Cannot parse value %s of '%s'.", type(nullable), nullable, property_name)

    return None


def to_nullable_double(nullable: Optional[Union[str, int, float]],
                       property_name: str, protos):
    """Converts int or float or str that parses to a number to an
    'NullableDouble' to be sent through the RPC layer. Input that is not a
    valid number but is also not null or undefined logs a function app level
    warning.
    :param nullable Input to be converted to an NullableDouble if it is a
    valid number
    :param property_name The name of the property that the caller will
    assign the output to. Used for debugging.
    :param protos The protos object used for returning the appropriate value
    """
    if isinstance(nullable, int) or isinstance(nullable, float):
        return protos.NullableDouble(value=nullable)
    elif isinstance(nullable, str):
        if len(nullable) == 0:
            return None

        try:
            return protos.NullableDouble(value=float(nullable))
        except Exception:
            raise TypeError(
                "Cannot parse value %s of '%s' to "
                "float.", nullable, property_name)

    if nullable is not None:
        raise TypeError(
            "A 'int' or 'float'"
            " type was expected instead of a '%s' "
            "type. Cannot parse value %s of '%s'.", type(nullable), nullable, property_name)

    return None


def to_nullable_timestamp(date_time: Optional[Union[datetime, int]],
                          property_name: str, protos):
    """Converts Date or number input to an 'NullableTimestamp' to be sent
    through the RPC layer. Input that is not a Date or number but is also
    not null or undefined logs a function app level warning.

    :param date_time Input to be converted to an NullableTimestamp if it is
    valid input
    :param property_name The name of the property that the caller will
    assign the output to. Used for debugging.
    :param protos The protos object used for returning the appropriate value
    """
    if date_time is not None:
        try:
            time_in_seconds = date_time if isinstance(date_time,
                                                      int) else \
                date_time.timestamp()

            return protos.NullableTimestamp(
                value=protos.Timestamp(seconds=int(time_in_seconds)))
        except Exception:
            raise TypeError(
                "A 'datetime' or 'int'"
                " type was expected instead of a '%s' "
                "type. Cannot parse value %s of '%s'.", type(date_time), date_time, property_name)
    return None



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from dataclasses import dataclass


class RpcException:
    def __init__(self,
                 source: str,
                 stack_trace: str,
                 message: str) -> None:
        self.__source = source
        self.__stack_trace = stack_trace
        self.__message = message

    @property
    def source(self) -> str:
        return self.__source

    @property
    def stack_trace(self) -> str:
        return self.__stack_trace

    @property
    def message(self) -> str:
        return self.__message


@dataclass
class RetryContext:
    """Gets the current retry count from retry-context"""
    retry_count: int
    """Gets the max retry count from retry-context"""
    max_retry_count: int
    rpc_exception: RpcException



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import os

from .logging import logger

from .utils.env_state import get_app_setting
from .utils.constants import (APPLICATIONINSIGHTS_CONNECTION_STRING,
                              PYTHON_AZURE_MONITOR_LOGGER_NAME,
                              PYTHON_AZURE_MONITOR_LOGGER_NAME_DEFAULT,
                              TRACESTATE, TRACEPARENT)


class OTelManager:
    def __init__(self):
        self._azure_monitor_available = False
        self._context_api = None
        self._trace_context_propagator = None

    def set_azure_monitor_available(self, azure_monitor_available):
        self._azure_monitor_available = azure_monitor_available

    def get_azure_monitor_available(self):
        return self._azure_monitor_available

    def set_context_api(self, context_api):
        self._context_api = context_api

    def get_context_api(self):
        return self._context_api

    def set_trace_context_propagator(self, trace_context_propagator):
        self._trace_context_propagator = trace_context_propagator

    def get_trace_context_propagator(self):
        return self._trace_context_propagator


def update_opentelemetry_status():
    """Check for OpenTelemetry library availability and
    update the status attribute."""
    try:
        from opentelemetry import context as context_api
        from opentelemetry.trace.propagation.tracecontext import (
            TraceContextTextMapPropagator,
        )

        OTelManager.set_context_api(context_api)
        OTelManager.set_trace_context_propagator(TraceContextTextMapPropagator())

    except ImportError:
        logger.exception(
            "Cannot import OpenTelemetry libraries."
        )


def initialize_azure_monitor():
    """Initializes OpenTelemetry and Azure monitor distro
    """
    update_opentelemetry_status()
    try:
        from azure.monitor.opentelemetry import configure_azure_monitor

        # Set functions resource detector manually until officially
        # include in Azure monitor distro
        os.environ.setdefault(
            "OTEL_EXPERIMENTAL_RESOURCE_DETECTORS",
            "azure_functions",
        )

        configure_azure_monitor(
            # Connection string can be explicitly specified in Appsetting
            # If not set, defaults to env var
            # APPLICATIONINSIGHTS_CONNECTION_STRING
            connection_string=get_app_setting(
                setting=APPLICATIONINSIGHTS_CONNECTION_STRING
            ),
            logger_name=get_app_setting(
                setting=PYTHON_AZURE_MONITOR_LOGGER_NAME,
                default_value=PYTHON_AZURE_MONITOR_LOGGER_NAME_DEFAULT
            ),
        )
        OTelManager.set_azure_monitor_available(True)

        logger.info("Successfully configured Azure monitor distro.")
    except ImportError:
        logger.exception(
            "Cannot import Azure Monitor distro."
        )
        OTelManager.set_azure_monitor_available(False)
    except Exception:
        logger.exception(
            "Error initializing Azure monitor distro."
        )
        OTelManager.set_azure_monitor_available(False)


def configure_opentelemetry(invocation_context):
    carrier = {TRACEPARENT: invocation_context.trace_context.trace_parent,
               TRACESTATE: invocation_context.trace_context.trace_state}
    ctx = OTelManager.get_trace_context_propagator().extract(carrier)
    OTelManager.get_context_api().attach(ctx)


otel_manager = OTelManager()



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

from .handle_event import worker_init_request, functions_metadata_request, function_environment_reload_request, invocation_request, function_load_request

__all__ = ('worker_init_request', 'functions_metadata_request', 'function_environment_reload_request', 'invocation_request', 'function_load_request')



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import inspect
import operator
import typing

from .bindings.meta import (has_implicit_output,
                            check_output_type_annotation,
                            check_input_type_annotation)
from .utils.constants import HTTP_TRIGGER
from .utils.typing_inspect import is_generic_type, get_origin, get_args


class ParamTypeInfo(typing.NamedTuple):
    binding_name: str
    pytype: typing.Optional[type]


class FunctionInfo(typing.NamedTuple):
    func: typing.Callable

    name: str
    directory: str
    function_id: str
    requires_context: bool
    is_async: bool
    has_return: bool
    is_http_func: bool

    input_types: typing.Mapping[str, ParamTypeInfo]
    output_types: typing.Mapping[str, ParamTypeInfo]
    return_type: typing.Optional[ParamTypeInfo]

    trigger_metadata: typing.Optional[typing.Dict[str, typing.Any]]


class FunctionLoadError(RuntimeError):

    def __init__(self, function_name: str, msg: str) -> None:
        super().__init__(
            "cannot load the " + function_name + " function: " + msg)


class Registry:
    _functions: typing.MutableMapping[str, FunctionInfo]

    def __init__(self) -> None:
        self._functions = {}

    def get_function(self, function_id: str) -> FunctionInfo:
        if function_id in self._functions:
            return self._functions[function_id]

        return None

    @staticmethod
    def get_explicit_and_implicit_return(binding_name: str,
                                         binding,
                                         explicit_return: bool,
                                         implicit_return: bool,
                                         bound_params: dict) -> \
            typing.Tuple[bool, bool]:
        if binding_name == '$return':
            explicit_return = True
        elif has_implicit_output(binding.type):
            implicit_return = True
            bound_params[binding_name] = binding
        else:
            bound_params[binding_name] = binding
        return explicit_return, implicit_return

    @staticmethod
    def get_return_binding(binding_name: str,
                           binding_type: str,
                           return_binding_name: str,
                           explicit_return_val_set: bool) \
            -> typing.Tuple[str, bool]:
        # prioritize explicit return value
        if explicit_return_val_set:
            return return_binding_name, explicit_return_val_set
        if binding_name == "$return":
            return_binding_name = binding_type
            assert return_binding_name is not None
            explicit_return_val_set = True
        elif has_implicit_output(binding_type):
            return_binding_name = binding_type

        return return_binding_name, explicit_return_val_set

    @staticmethod
    def validate_binding_direction(binding_name: str,
                                   binding_direction: str,
                                   func_name: str,
                                   protos):
        if binding_direction == protos.BindingInfo.inout:
            raise FunctionLoadError(
                func_name,
                '"inout" bindings are not supported')

        if binding_name == '$return' and \
                binding_direction != protos.BindingInfo.out:
            raise FunctionLoadError(
                func_name,
                '"$return" binding must have direction set to "out"')

    @staticmethod
    def is_context_required(params, bound_params: dict,
                            annotations: dict,
                            func_name: str) -> bool:
        requires_context = False
        if 'context' in params and 'context' not in bound_params:
            requires_context = True
            params.pop('context')
            if 'context' in annotations:
                ctx_anno = annotations.get('context')
                if (not isinstance(ctx_anno, type)
                        or ctx_anno.__name__ != 'Context'):
                    raise FunctionLoadError(
                        func_name,
                        'the "context" parameter is expected to be of '
                        'type azure.functions.Context, got "%s"', repr(ctx_anno))
        return requires_context

    @staticmethod
    def validate_function_params(params: dict, bound_params: dict,
                                 annotations: dict, func_name: str,
                                 protos):
        if set(params) - set(bound_params):
            raise FunctionLoadError(
                func_name,
                'the following parameters are declared in Python but '
                'not in function.json: %s', repr(set(params) - set(bound_params)))

        if set(bound_params) - set(params):
            raise FunctionLoadError(
                func_name,
                'the following parameters are declared in function.json but '
                'not in Python: %s', repr(set(params) - set(bound_params)))

        input_types: typing.Dict[str, ParamTypeInfo] = {}
        output_types: typing.Dict[str, ParamTypeInfo] = {}

        for param in params.values():
            binding = bound_params[param.name]

            param_has_anno = param.name in annotations
            param_anno = annotations.get(param.name)

            if param_has_anno:
                if is_generic_type(param_anno):
                    param_anno_origin = get_origin(param_anno)
                    if param_anno_origin is not None:
                        is_param_out = (
                            isinstance(param_anno_origin, type)
                            and param_anno_origin.__name__ == 'Out'
                        )
                    else:
                        is_param_out = (
                            isinstance(param_anno, type)
                            and param_anno.__name__ == 'Out'
                        )
                else:
                    is_param_out = (
                        isinstance(param_anno, type)
                        and param_anno.__name__ == 'Out'
                    )
            else:
                is_param_out = False

            is_binding_out = binding.direction == protos.BindingInfo.out

            if is_param_out:
                param_anno_args = get_args(param_anno)
                if len(param_anno_args) != 1:
                    raise FunctionLoadError(
                        func_name,
                        'binding %s has invalid Out annotation %s', param.name, repr(param_anno))
                param_py_type = param_anno_args[0]

                # typing_inspect.get_args() returns a flat list,
                # so if the annotation was func.Out[typing.List[foo]],
                # we need to reconstruct it.
                if (isinstance(param_py_type, tuple)
                        and is_generic_type(param_py_type[0])):
                    param_py_type = operator.getitem(
                        param_py_type[0], *param_py_type[1:])
            else:
                param_py_type = param_anno

            if (param_has_anno and not isinstance(param_py_type, type)
                    and not is_generic_type(param_py_type)):
                raise FunctionLoadError(
                    func_name,
                    'binding %s has invalid non-type annotation %s', param.name, repr(param_anno))

            if is_binding_out and param_has_anno and not is_param_out:
                raise FunctionLoadError(
                    func_name,
                    'binding %s is declared to have the "out" '
                    'direction, but its annotation in Python is not '
                    'a subclass of azure.functions.Out', param.name)

            if not is_binding_out and is_param_out:
                raise FunctionLoadError(
                    func_name,
                    'binding %s is declared to have the "in" '
                    'direction in function.json, but its annotation '
                    'is azure.functions.Out in Python', param.name)

            if param_has_anno and param_py_type in (str, bytes) and (
                    not has_implicit_output(binding.type)):
                param_bind_type = 'generic'
            else:
                param_bind_type = binding.type

            if param_has_anno:
                if is_param_out:
                    checks_out = check_output_type_annotation(
                        param_bind_type, param_py_type)
                else:
                    checks_out = check_input_type_annotation(
                        param_bind_type, param_py_type)

                if not checks_out:
                    if binding.data_type is not protos.BindingInfo.undefined:
                        raise FunctionLoadError(
                            func_name,
                            '%s binding type "%s" '
                            'and dataType "%s" in '
                            'function.json do not match the corresponding '
                            'function parameter\'s Python type '
                            'annotation %s', repr(param.name), binding.type, binding.data_type, param_py_type.__name__)
                    else:
                        raise FunctionLoadError(
                            func_name,
                            'type of %s binding in function.json '
                            '"%s" does not match its Python '
                            'annotation "%s"', param.name, binding.type, param_py_type.__name__)

            param_type_info = ParamTypeInfo(param_bind_type,
                                            param_py_type)
            if is_binding_out:
                output_types[param.name] = param_type_info
            else:
                input_types[param.name] = param_type_info
        return input_types, output_types

    @staticmethod
    def get_function_return_type(annotations: dict, has_explicit_return: bool,
                                 has_implicit_return: bool, binding_name: str,
                                 func_name: str):
        return_pytype = None
        if has_explicit_return and 'return' in annotations:
            return_anno = annotations.get('return')
            if is_generic_type(
                    return_anno) and get_origin(
                    return_anno).__name__ == 'Out':
                raise FunctionLoadError(
                    func_name,
                    'return annotation should not be azure.functions.Out')

            return_pytype = return_anno
            if not isinstance(return_pytype, type):
                raise FunctionLoadError(
                    func_name,
                    'has invalid non-type return '
                    'annotation %s', repr(return_pytype))

            if return_pytype is (str, bytes):
                binding_name = 'generic'

            if not check_output_type_annotation(
                    binding_name, return_pytype):
                raise FunctionLoadError(
                    func_name,
                    'Python return annotation "%s" '
                    'does not match binding type "%s"', return_pytype.__name__, binding_name)

        if has_implicit_return and 'return' in annotations:
            return_pytype = annotations.get('return')

        return_type = None
        if has_explicit_return or has_implicit_return:
            return_type = ParamTypeInfo(binding_name, return_pytype)

        return return_type

    def add_func_to_registry_and_return_funcinfo(
            self, function,
            function_name: str,
            function_id: str,
            directory: str,
            requires_context: bool,
            has_explicit_return: bool,
            has_implicit_return: bool,
            input_types: typing.Dict[str, ParamTypeInfo],
            output_types: typing.Dict[str, ParamTypeInfo],
            return_type: str):

        http_trigger_param_name = self._get_http_trigger_param_name(input_types)

        trigger_metadata = None
        is_http_func = False
        if http_trigger_param_name is not None:
            trigger_metadata = {
                "type": HTTP_TRIGGER,
                "param_name": http_trigger_param_name
            }
            is_http_func = True

        function_info = FunctionInfo(
            func=function,
            name=function_name,
            directory=directory,
            function_id=function_id,
            requires_context=requires_context,
            is_async=inspect.iscoroutinefunction(function),
            has_return=has_explicit_return or has_implicit_return,
            is_http_func=is_http_func,
            input_types=input_types,
            output_types=output_types,
            return_type=return_type,
            trigger_metadata=trigger_metadata)

        self._functions[function_id] = function_info

        return function_info

    def _get_http_trigger_param_name(self, input_types):
        http_trigger_param_name = next(
            (input_type for input_type, type_info in input_types.items()
             if type_info.binding_name == HTTP_TRIGGER),
            None
        )
        return http_trigger_param_name

    def add_function(self, function_id: str,
                     func: typing.Callable,
                     metadata,
                     protos):
        func_name = metadata.name
        sig = inspect.signature(func)
        params = dict(sig.parameters)
        annotations = typing.get_type_hints(func)
        return_binding_name: typing.Optional[str] = None
        explicit_return_val_set = False
        has_explicit_return = False
        has_implicit_return = False

        bound_params = {}
        for binding_name, binding_info in metadata.bindings.items():
            self.validate_binding_direction(binding_name,
                                            binding_info.direction, func_name,
                                            protos)

            has_explicit_return, has_implicit_return = \
                self.get_explicit_and_implicit_return(
                    binding_name, binding_info, has_explicit_return,
                    has_implicit_return, bound_params)

            return_binding_name, explicit_return_val_set = \
                self.get_return_binding(binding_name,
                                        binding_info.type,
                                        return_binding_name,
                                        explicit_return_val_set)

        requires_context = self.is_context_required(params, bound_params,
                                                    annotations,
                                                    func_name)

        input_types, output_types = self.validate_function_params(
            params, bound_params, annotations, func_name, protos)

        return_type = \
            self.get_function_return_type(annotations,
                                          has_explicit_return,
                                          has_implicit_return,
                                          return_binding_name,
                                          func_name)

        self.add_func_to_registry_and_return_funcinfo(func, func_name,
                                                      function_id,
                                                      metadata.directory,
                                                      requires_context,
                                                      has_explicit_return,
                                                      has_implicit_return,
                                                      input_types,
                                                      output_types,
                                                      return_type)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
import os
import sys

from typing import Any, Dict, Optional

from .datumdef import Datum, datum_as_proto
from .generic import GenericBinding

from ..utils.constants import (
    CUSTOMER_PACKAGES_PATH,
)

PB_TYPE = 'rpc_data'
PB_TYPE_DATA = 'data'
PB_TYPE_RPC_SHARED_MEMORY = 'rpc_shared_memory'


def _check_http_input_type_annotation(bind_name: str, pytype: type) -> bool:
    binding = get_binding(bind_name)
    return binding.check_input_type_annotation(pytype)


def _check_http_output_type_annotation(bind_name: str, pytype: type) -> bool:
    binding = get_binding(bind_name)
    return binding.check_output_type_annotation(pytype)


def load_binding_registry() -> None:
    """
    Tries to load azure-functions from the customer's BYO. If it's
    not found, it loads the builtin. If the BINDING_REGISTRY is None,
    azure-functions hasn't been loaded in properly.

    Tries to load the base extension only for python 3.8+.
    """
    func = sys.modules.get('azure.functions')
    if func is None:
        import azure.functions as func

    global BINDING_REGISTRY
    BINDING_REGISTRY = func.get_binding_registry()

    if BINDING_REGISTRY is None:
        raise AttributeError('BINDING_REGISTRY is None. azure-functions '
                             'library not found. Sys Path: %s. '
                             'Sys Modules: %s. '
                             'python-packages Path exists: %s.',
                             sys.path, sys.modules,
                             os.path.exists(CUSTOMER_PACKAGES_PATH))


def get_binding(bind_name: str)\
        -> object:
    """
    First checks if the binding is not generic.
    If binding is still None, it's a generic type.
    """
    binding = BINDING_REGISTRY.get(bind_name)
    if binding is None:
        binding = GenericBinding
    return binding


def is_trigger_binding(bind_name: str) -> bool:
    binding = get_binding(bind_name)
    return binding.has_trigger_support()


def check_input_type_annotation(bind_name: str,
                                pytype: type) -> bool:
    binding = get_binding(bind_name)

    return binding.check_input_type_annotation(pytype)


def check_output_type_annotation(bind_name: str, pytype: type) -> bool:
    binding = get_binding(bind_name)
    return binding.check_output_type_annotation(pytype)


def has_implicit_output(bind_name: str) -> bool:
    binding = get_binding(bind_name)

    # Need to pass in bind_name to exempt Durable Functions
    if binding is GenericBinding:
        return (getattr(binding, 'has_implicit_output', lambda: False)
                (bind_name))

    else:
        # If the binding does not have metaclass of meta.InConverter
        # The implicit_output does not exist
        return getattr(binding, 'has_implicit_output', lambda: False)()


def from_incoming_proto(
        binding: str,
        pb, *,
        trigger_metadata: Optional[Dict[str, Any]]) -> Any:
    binding = get_binding(binding)
    if trigger_metadata:
        metadata = {
            k: Datum.from_typed_data(v)
            for k, v in trigger_metadata.items()
        }
    else:
        metadata = {}

    pb_type = pb.WhichOneof(PB_TYPE)
    if pb_type == PB_TYPE_DATA:
        val = pb.data
        datum = Datum.from_typed_data(val)
    else:
        raise TypeError('Unknown ParameterBindingType: %s', pb_type)

    try:
        return binding.decode(datum, trigger_metadata=metadata)
    except NotImplementedError:
        # Binding does not support the data.
        dt = val.WhichOneof('data')
        raise TypeError(
            'unable to encode outgoing TypedData: '
            'unsupported type "%s" for '
            'Python type "%s"', binding, type(obj).__name__)


def get_datum(binding: str, obj: Any,
              pytype: Optional[type]) -> Datum:
    """
    Convert an object to a datum with the specified type.
    """
    binding = get_binding(binding)
    try:
        datum = binding.encode(obj, expected_type=pytype)
    except NotImplementedError:
        # Binding does not support the data.
        raise TypeError(
            'unable to encode outgoing TypedData: '
            'unsupported type "%s" for '
            'Python type "%s"', binding, type(obj).__name__)
    return datum


def _does_datatype_support_caching(datum: Datum):
    supported_datatypes = ('bytes', 'string')
    return datum.type in supported_datatypes


def to_outgoing_proto(binding: str, obj: Any, *,
                      pytype: Optional[type],
                      protos):
    datum = get_datum(binding, obj, pytype)
    return datum_as_proto(datum, protos)


def to_outgoing_param_binding(binding: str, obj: Any, *,
                              pytype: Optional[type],
                              out_name: str,
                              protos):
    datum = get_datum(binding, obj, pytype)
    # If not, send it as part of the response message over RPC
    # rpc_val can be None here as we now support a None return type
    rpc_val = datum_as_proto(datum, protos)
    return protos.ParameterBinding(
        name=out_name,
        data=rpc_val)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

import logging.handlers
import sys
import traceback

# Logging Prefixes
SDK_LOG_PREFIX = "azure.functions"

logger: logging.Logger = logging.getLogger(SDK_LOG_PREFIX)


def format_exception(exception: Exception) -> str:
    msg = str(exception) + "\n"
    if (sys.version_info.major, sys.version_info.minor) < (3, 10):
        msg += ''.join(traceback.format_exception(
            etype=type(exception),
            tb=exception.__traceback__,
            value=exception))
    elif (sys.version_info.major, sys.version_info.minor) >= (3, 10):
        msg += ''.join(traceback.format_exception(exception))
    else:
        msg = str(exception)
    return msg
